#include "jserver.h"

thread session_create(jid owner, session_onSend evt, void *evt_data)
{
    /* creats a new session, registers the resource for it, and spawns a thread to handle it */
    thread t;
    session s;
    xmlnode presence;
    jid user;
    pth_attr_t attr;

    if(owner == NULL || owner->resource == NULL) /* illegal */
        return NULL;
 
    jdebug(ZONE,"session_create");
 
    /* create session thread */
    t = thread_new();
    s = pmalloc(t->p, sizeof(struct session_struct));
    t->data = (void *)s;

    /* setup internal communication points */
    s->mp_from = pth_msgport_create("session_from");
    s->mp_to = pth_msgport_create("session_to");
    s->send_evt = evt;
    s->send_evt_data = evt_data;

    /* session identity and default settings */
    s->id = jid_new(t->p, jid_full(owner));
    s->r = resource_create(RESOURCE_SESSION, owner->user, owner->resource, s->mp_to);
    s->r->aux = (void *)s;
    s->p = t->p;
    s->in_count = s->out_count = s->exit_flag = 0;
    s->presence = xmlnode_new_tag("presence");
    xmlnode_put_attrib(s->presence,"type","unavailable");

    /* only for non-anonymous users */
    if(owner->user != NULL)
    {
        s->u = user_data(owner->user);
        if(s->u->primary != NULL)
        {
            /* we'll be default since we're the most recent */
            if(s->u->priority == 0)
            {
                jdebug(ZONE,"MASTER %s",s->id->resource);
                s->u->primary = s->r;
                s->u->priority = 0;
            }

            /* built-in presence broadcast to any of this users other sessions only */
            presence = xmlnode_new_tag("presence");
            user = jid_new(t->p,jid_full(owner));
            jid_set(user,NULL,JID_RESOURCE); /* broadcast */
            xmlnode_put_attrib(presence,"to",jid_full(user));
            xmlnode_put_attrib(presence,"from",jid_full(owner));
            xmlnode_put_attrib(presence,"type","available");
            deliver(jpacket_new(presence));
        }else{
            jdebug(ZONE,"MASTER %s",s->id->resource);
            s->u->primary = s->r;
            s->u->priority = 0;
        }
    }else{
        s->u = NULL;
    }

    /* start thread */
    attr = pth_attr_new();
    pth_attr_set(attr, PTH_ATTR_JOINABLE, FALSE);
    pth_attr_set(attr, PTH_ATTR_STACK_SIZE, jlib_stack_default);
    t->id = pth_spawn(attr, session_main, (void *)t);
    pth_attr_destroy(attr);
    pth_yield(NULL); /* let it have a chance to start up */
 
    return t;
}

void session_kill(char *user, char *res, char *reason)
{
    resource r;

    r = resource_lookup(user, res);
    if(r == NULL || r->type != RESOURCE_SESSION)
        return;

    if(reason != NULL)
        session_end((session)(r->aux), reason);
    else
        session_end((session)(r->aux), "killed");
}

void session_end(session s, char *reason)
{
    xmlnode x;
    jid to_id;

    if(s == NULL || s->exit_flag == 1 || reason == NULL)
        return;

    jdebug(ZONE,"end %d '%s'",s,reason);

    /* flag the session to exit ASAP */
    s->exit_flag = 1;

    /* make sure we're not the primary session */
    if(s->u != NULL && s->u->primary == s->r)
    {
        jdebug(ZONE,"UN-MASTER %s",s->id->resource);
        s->u->primary = NULL;
        s->u->priority = 0;
    }

    /* create a presence packet for this action */
    x = xmlnode_new_tag("presence");
    xmlnode_put_attrib(x,"type","unavailable");
    xmlnode_insert_cdata(xmlnode_insert_tag(x,"status"),reason,strlen(reason));

    /* if this session still thinks it's online, tell it otherwise */
    if(j_strcmp(xmlnode_get_attrib(s->presence,"type"),"unavailable") != 0)
    {
        psend(s->mp_from,jpacket_new(x));
    }else{
        /* it was already listed as unavailable, but that might be since it never set any presence, 
         * we still need to internally tell the user's other resources that it's gone */
        if(s->r->user != NULL)
        {
            xmlnode_put_attrib(x,"from",jid_full(s->id));
            to_id = jid_new(xmlnode_pool(x),jid_full(s->id));
            jid_set(to_id,NULL,JID_RESOURCE); /* broadcast */
            xmlnode_put_attrib(x,"to",jid_full(to_id));
            deliver(jpacket_new(xmlnode_dup(x)));
        }
        /* send this as a dummy presence so that the session at least gets CPU time */
        psend(s->mp_to,jpacket_new(x));
    }
}

void *session_main(void *arg)
{
    thread t = (thread)arg;
    session s = (session)(t->data);
    pth_event_t from, to, ring;
    jpq q;

    jdebug(ZONE,"THREAD:SESSION starting");

    from = pth_event(PTH_EVENT_MSG,s->mp_from);
    to = pth_event(PTH_EVENT_MSG,s->mp_to);
    ring = pth_event_concat(from, to, NULL);

    module_dispatch(mapi_new(M_PHASE_SESSION_ON,t->p,s,NULL));

    while(1)
    {
        pth_wait(ring);

        while(1)
        {
            /* this loop looks for messages from our user that need to be delivered elsewhere */
            q = (jpq)pth_msgport_get(s->mp_from);
            if(q == NULL)
                break;

            jdebug(ZONE,"THREAD:SESSION:FROM received a message!");
            ++s->out_count;

            if(s->r->user != NULL)
                if(session_from_handle(s,q->p) == 1)
                    continue;

            /* make sure we have our from set correctly for outgoing crap */
            xmlnode_put_attrib(q->p->x,"from",jid_full(s->id));
            q->p->from = jid_new(q->p->p,jid_full(s->id));

            /* pass these to the general delivery function */
            deliver(q->p);
        }

        if(s->exit_flag)
            break;

        while(1)
        {
            /* this loop looks for messages to our user that need to be sent to them */
            q = (jpq)pth_msgport_get(s->mp_to);
            if(q == NULL)
                break;

            jdebug(ZONE,"THREAD:SESSION:TO received data from %s!",jid_full(q->p->from));
            ++s->in_count;

            if(s->r->user != NULL)
                if(session_to_handle(s,q->p) == 1)
                    continue;

            /* we need to check again, s->exit_flag *could* have changed within session_to_handle at some point */
            if(s->exit_flag)
            {
                bounce(q->p,"Session Disconnected");
                continue;
            }

            /* deliver outgoing for this session to the onSend event passed when this session was created */
            (s->send_evt)(t, q->p);
        }

    }

    jdebug(ZONE,"THREAD:SESSION exiting");

    /* unlist ourselves */
    resource_free(s->r);

    /* make sure the service knows the session is gone */
    (s->send_evt)(t, NULL);

    module_dispatch(mapi_new(M_PHASE_SESSION_OFF,t->p,s,NULL));

    xmlnode_free(s->presence);
    pth_event_free(ring,PTH_FREE_ALL);
    pth_msgport_destroy(s->mp_from);

    /* the delivery waiting packets, if any */
    while(1)
    {
        q = (jpq)pth_msgport_get(s->mp_to);
        if(q == NULL)
            break;
        if(q->p->type == JPACKET_MESSAGE)
            deliver(q->p); /* re-deliver the messages only */
        else
            xmlnode_free(q->p->x);
    }
    pth_msgport_destroy(s->mp_to);

    logger_session(s);

    thread_free(t);

    return NULL;
}


int session_from_handle(session s, jpacket p)
{
    xmlnode presence;
    mapi m;
    jid to_id;
    int flag;

                switch(p->type)
                {
                    case JPACKET_MESSAGE:
                        m = mapi_new(M_PHASE_MESSAGE_OUT,p->p,s,p);
                        module_dispatch(m);
                        if(m->result != M_DEFAULT)
                        {
                            xmlnode_free(p->x);
                            return 1;
                        }
                        break;

                    case JPACKET_PRESENCE:
                        /* it was a target send */
                        if(p->to != NULL)
                            return 0;

                        /* ha ha, we're better than you, give us default messages! */
                        if(presence_priority(p->x) > s->u->priority)
                        {
                            jdebug(ZONE,"MASTER %s",s->id->resource);
                            s->u->priority = presence_priority(p->x);
                            s->u->primary = s->r;
                        }else if(s->u->primary == s->r){ /* we were default, but we're lower now :( */
                            s->u->priority = presence_priority(p->x);
                        }

                        /* force broadcast to our other resources */
                        presence = xmlnode_dup(p->x);
                        xmlnode_put_attrib(presence,"from",jid_full(s->id));
                        to_id = jid_new(p->p,jid_full(s->id));
                        jid_set(to_id,NULL,JID_RESOURCE);
                        xmlnode_put_attrib(presence,"to",jid_full(to_id));
                        deliver(jpacket_new(presence));

                        m = mapi_new(M_PHASE_PRESENCE_OUT,p->p,s,p);
                        module_dispatch(m);
                        if(m->result == M_DEFAULT)
                        { /* module didn't handle it, we'll use our internal roster */
                            to_id = s->u->subscribers;
                            flag = j_strcmp(xmlnode_get_attrib(p->x,"type"),"available");
                            if(flag != 0)
                                flag = j_strcmp(xmlnode_get_attrib(p->x,"type"),"unavailable");
                            while(to_id != NULL)
                            {
                                if(flag != 0 && presence_check(CMD_GET,jid_full(to_id)) == 0)
                                { /* if this isn't an [un]available type, only send to those we know are available */
                                    to_id = to_id->next;
                                    continue;
                                }
                                jid_set(to_id,NULL,JID_RESOURCE);
                                presence = xmlnode_dup(p->x);
                                xmlnode_put_attrib(presence,"from",jid_full(s->id));
                                xmlnode_put_attrib(presence,"to",jid_full(to_id));
                                deliver(jpacket_new(presence));
                                to_id = to_id->next;
                            }
                        }

                        /* replace this sessions current status */
                        xmlnode_free(s->presence);
                        s->presence = p->x;
                        if(j_strcmp(xmlnode_get_attrib(s->presence,"type"),"available") == 0)
                            xmlnode_hide_attrib(s->presence,"type");
                        return 1;

                    case JPACKET_IQ:
                        /* it was a target send */
                        if(p->to != NULL)
                            return 0;

                        /* known IQ handlers or let module handle it */
                        if(j_strcmp(xmlnode_get_attrib(xmlnode_get_tag(p->x,"query"),"xmlns"),"jabber:iq:roster") == 0)
                            session_roster_handle(s,p);
                        else
                            module_dispatch(mapi_new(M_PHASE_IQ_HANDLE,p->p,s,p));

                        /* reply */
                        iq_reply(p->x);
                        psend(s->mp_to, p);
                        return 1;
                }

    return 0;
}


int session_to_handle(session s, jpacket p)
{
    xmlnode presence;
    mapi m;
    jid to_id;
    int ret = 0;

                switch(p->type)
                {

                    case JPACKET_MESSAGE:
                        m = mapi_new(M_PHASE_MESSAGE_IN,p->p,s,p);
                        module_dispatch(m);
                        if(m->result != M_DEFAULT)
                            ret = 1;
                        break;

                    case JPACKET_PRESENCE:
                        /* we don't want to talk to ourselves :) */
                        if(jid_cmp(p->from,s->id) == 0)
                        {
                            ret = 1;
                            break;
                        }

                        /* one of our sessions is changing presence, internal broadcast */
                        if(j_strcmp(p->from->user, s->id->user) == 0 && config_names(CMD_CHECK,p->from->server) != NULL)
                        {
                            /* if they lowered, recheck the default in case were better */
                            if(presence_priority(s->presence) > s->u->priority || s->u->primary == NULL)
                            {
                                jdebug(ZONE,"MASTER %s",s->id->resource);
                                s->u->priority = presence_priority(s->presence);
                                s->u->primary = s->r;
                            }
                            /* reply */
                            if(j_strcmp(xmlnode_get_attrib(p->x,"type"),"available") == 0 || j_strcmp(xmlnode_get_attrib(p->x,"type"),"poll") == 0)
                            {
                                presence = xmlnode_dup(s->presence);
                                xmlnode_put_attrib(presence,"to",jid_full(p->from));
                                xmlnode_put_attrib(presence,"from",jid_full(s->id));
                                deliver(jpacket_new(presence));
                            }
                            break;
                        }

                        m = mapi_new(M_PHASE_PRESENCE_IN,p->p,s,p);
                        module_dispatch(m);
                        if(m->result == M_DEFAULT)
                        { /* the module didn't handle responding, we will */
                            if(j_strcmp(xmlnode_get_attrib(p->x,"type"),"available") == 0 || j_strcmp(xmlnode_get_attrib(p->x,"type"),"poll") == 0)
                            {
                                to_id = s->u->subscribers;
                                while(to_id != NULL)
                                {
                                    if(jid_cmp(p->from,to_id) == 0)
                                        break;
                                    to_id = to_id->next;
                                }
                                if(to_id != NULL)
                                { /* only reply if they are on our subscription */
                                    presence = xmlnode_dup(s->presence);
                                    xmlnode_put_attrib(presence,"to",jid_full(p->from));
                                    xmlnode_put_attrib(presence,"from",jid_full(s->id));
                                    deliver(jpacket_new(presence));
                                }
                                /* polls don't get delivered */
                                if(j_strcmp(xmlnode_get_attrib(p->x,"type"),"poll") == 0)
                                    ret = 1;
                            }else if(j_strcmp(xmlnode_get_attrib(p->x,"type"),"subscribe") == 0){
                                /* XXX options on the account, autosubscribe? */
                            }
                        }
                        break;
                    case JPACKET_IQ:
                        /* internal response IQ from mp_from */
                        if(p->to == NULL)
                            break;

                        m = mapi_new(M_PHASE_IQ,p->p,s,p);
                        module_dispatch(m);
                        if(m->result == M_HANDLED)
                        {
                            /* return the result to the sender */
                            iq_reply(p->x);
                            deliver(jpacket_new(p->x));
                        }
                        if(m->result != M_DEFAULT) /* send to user unless the module took care of it */
                            ret = 1;
                        break;
                }

    if(ret)
        xmlnode_free(p->x);

    return ret;
}

